@@ -1,11 +1,15 @@
+import asyncio
+import functools
 import logging
 import sqlite3
-from collections.abc import Generator
+from collections.abc import Callable, Generator
 from contextlib import contextmanager
 from datetime import UTC, datetime
 from pathlib import Path
 from typing import Self
 
+from opentelemetry import context
+
 from .. import RateLimitingRepo, Usage
 
 _LOG = logging.getLogger(__name__)
@@ -24,11 +28,27 @@ class SqliteRateLimitingRepo(RateLimitingRepo):
 
         self._connection = connection
 
+    @staticmethod
+    async def _run_in_executor[T](func: Callable[[], T]) -> T:
+        loop = asyncio.get_running_loop()
+        ctx = context.get_current()
+
+        def __with_context() -> T:
+            token = context.attach(ctx)
+            try:
+                return func()
+            finally:
+                context.detach(token)
+
+        return await loop.run_in_executor(None, __with_context)
+
     @classmethod
-    def connect(cls, db_file: Path) -> Self:
+    async def connect(cls, db_file: Path) -> Self:
         if db_file.exists() and not db_file.is_file():
             raise ValueError(f"Database file {db_file} exists and is not a file")
-        connection = sqlite3.connect(db_file)
+
+        partial = functools.partial(sqlite3.connect, db_file)
+        connection = await cls._run_in_executor(partial)
         return cls(connection)
 
     @contextmanager
@@ -40,7 +60,26 @@ class SqliteRateLimitingRepo(RateLimitingRepo):
         finally:
             cursor.close()
 
-    def add_usage(
+    async def add_usage(
+        self,
+        *,
+        context_id: str,
+        user_id: str,
+        utc_time: datetime,
+        reference_id: str | None,
+        response_id: str | None,
+    ):
+        partial = functools.partial(
+            self._add_usage,
+            context_id=context_id,
+            user_id=user_id,
+            utc_time=utc_time,
+            reference_id=reference_id,
+            response_id=response_id,
+        )
+        await self._run_in_executor(partial)
+
+    def _add_usage(
         self,
         *,
         context_id: str,
@@ -71,7 +110,22 @@ class SqliteRateLimitingRepo(RateLimitingRepo):
             )
         _LOG.debug("Inserted usage for user %s in context %s", user_id, context_id)
 
-    def get_usages(
+    async def get_usages(
+        self,
+        *,
+        context_id: str,
+        user_id: str,
+        limit: int = 1,
+    ) -> list[Usage]:
+        partial = functools.partial(
+            self._get_usages,
+            context_id=context_id,
+            user_id=user_id,
+            limit=limit,
+        )
+        return await self._run_in_executor(partial)
+
+    def _get_usages(
         self,
         *,
         context_id: str,
@@ -109,7 +163,11 @@ class SqliteRateLimitingRepo(RateLimitingRepo):
 
         return usages
 
-    def drop_old_usages(self, *, until: datetime) -> None:
+    async def drop_old_usages(self, *, until: datetime) -> None:
+        partial = functools.partial(self._drop_old_usages, until=until)
+        await self._run_in_executor(partial)
+
+    def _drop_old_usages(self, *, until: datetime) -> None:
         with self._cursor() as cursor:
             cursor.execute(
                 """
@@ -119,5 +177,5 @@ class SqliteRateLimitingRepo(RateLimitingRepo):
                 [until.timestamp()],
             )
 
-    def close(self) -> None:
-        self._connection.close()
+    async def close(self) -> None:
+        await self._run_in_executor(self._connection.close)
